/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/entity.h"
#include "caosdb/entity/v1alpha1/main.pb.h" // for Parent, Arena::CreateMay...
#include "caosdb/protobuf_helper.h"         // for get_arena
#include "google/protobuf/arena.h"          // for Arena

namespace caosdb::entity {
using caosdb::entity::v1alpha1::IdResponse;
using ProtoParent = caosdb::entity::v1alpha1::Parent;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using caosdb::utility::get_arena;

Parent::Parent() : wrapped(Parent::CreateProtoParent()) {}

auto Parent::CreateProtoParent() -> ProtoParent * {
  return google::protobuf::Arena::CreateMessage<ProtoParent>(get_arena());
}

auto Parent::SetName(const std::string &name) -> void {
  this->wrapped->set_name(name);
}

auto Parent::SetId(const std::string &id) -> void { this->wrapped->set_id(id); }

[[nodiscard]] auto Parent::GetId() const -> const std::string & {
  return this->wrapped->id();
}

[[nodiscard]] auto Parent::GetName() const -> const std::string & {
  return this->wrapped->name();
}

auto Parents::Append(const Parent &parent) -> void {
  auto *destination = this->wrapped->Add();
  destination->Swap(parent.wrapped);

  // Clear the originally wrapped object and return it to the Arena
  parent.wrapped->Clear();

  // set the pointer to the new object which is owned by the RepeatedPtrField
  parent.wrapped = destination;
}

[[nodiscard]] auto Entity::GetParents() const -> const Parents & {
  return parents;
}

auto Entity::AppendParent(const Parent &parent) -> void {
  this->parents.Append(parent);
}

auto Entity::CreateProtoEntity() -> ProtoEntity * {
  return google::protobuf::Arena::CreateMessage<ProtoEntity>(get_arena());
}

Entity::Entity() : wrapped(Entity::CreateProtoEntity()) {
  properties.wrapped = this->wrapped->mutable_properties();
  parents.wrapped = this->wrapped->mutable_parents();
  errors.wrapped = this->wrapped->mutable_errors();
}

Entity::Entity(IdResponse *idResponse) : Entity() {
  this->wrapped->set_id(idResponse->id());
  this->wrapped->mutable_errors()->Swap(idResponse->mutable_entity_errors());
}

auto Entity::SetId(const std::string &id) -> void { this->wrapped->set_id(id); }

auto Entity::SetVersionId(const std::string &id) -> void {
  this->wrapped->mutable_version()->set_id(id);
}

auto Entity::Switch(ProtoEntity *entity) -> void {
  this->wrapped->Swap(entity);
  this->wrapped->Clear();
  this->wrapped = entity;
}

auto Entity::SetRole(const std::string &role) -> void {
  this->wrapped->set_role(role);
}

auto Entity::SetName(const std::string &name) -> void {
  this->wrapped->set_name(name);
}

} // namespace caosdb::entity
